/*
 bug ????
 cross-validation also somehow seems to throw off the similarity computation
*/

#include <stdio.h>
#include "ripper.h"
#include "protos.h"
#include "mdb.h"

static double binomial_std_err(double,int);
static double rocchio_error_rate(vec_t *,char *,BOOL);
static void model_rocchio(vec_t *);
static int eval_final_rule(vec_t *);
static double eval_rule(vec_t *);
static double inner_product(vec_t *,double *);
static void alloc_wts();

/* adjustable parameters */
static double Beta;
static double Gamma;
static BOOL Transform_data;
static BOOL Two_prototypes;

/* what to optimize */
static enum {ACCURACY,F1} Metric;
BOOL Predict;
BOOL Predict_probabilities;

/* for probabilistic prediction, where one
 * outputs a probability 0-1 */
static double Max_sim;

#define words(exi) (vref(aval_t,exi->inst,0)->u.set)

/* for transforming data */
static double *Sum_wt,*Sum_wt_p,*Sum_wt_n;


/* interface to cross-validation
*/

static void train_rocchio(vec_t *data)
{
    model_rocchio(data);
}
static double test_rocchio(vec_t *data)
{
    return rocchio_error_rate(data,"foobar",FALSE);
}

/******************************************************************************/
char *Program="rocchio";
char *Help_str[] = {
    "syntax: rocchio [options] [stem]",
    "   use Rocchio's method to learn a classifier",
    "",
    "options are:",
    "  -s:       read from std input",
    "  -kN       estimate error rate by N-fold cross-validation",   
    "  -l        estimate error rate via leave-one-out method",
    "  -Ln       set loss ratio",
    "  -2        use two prototype vectors",
    "  -Pn       set factor for positive prototype (default 16)",
    "  -Nn       set factor for positive prototype (default 4)",
    "  -v#:      set verbosity",
    "  -m s:     set threshold to optimize metric s",
    "              s must be 'accuracy' or 'f-measure' (default 'accuracy')",
    "  -p:       print predictions for test cases to stem.pred",
    "  -X:       transform data by adding distance to prototype",
    "             (creates stem_cos.data, stem_cos.test)",
    "  -R:       rank test cases with 'probabilistic' predictions",
    NULL
};


main(argc,argv)
int argc;
char *argv[];
{
    vec_t *train_data;
    vec_t *test_data;
    char *stem=NULL;
    BOOL use_stdin;
    int o;
    int folds;
    double err; 
    BOOL crossval;
    double loss_ratio;
    example_t *exi;
    int i;
    FILE *fptest,*fptrain;
    
    crossval = FALSE;
    Class_ordering = INCFREQ;

    /* defaults */
    use_stdin = FALSE;
    set_trace_level(SUMM);
    Metric = ACCURACY;
    Predict = FALSE;
    FP_cost = FN_cost = 1.0;
    Predict_probabilities = FALSE;
    Two_prototypes = FALSE;

    Beta = 16.0;
    Gamma = 4.0;

    while ((o=getopt(argc,argv,"stv:m:pP:N:k:lL:hRX2"))!=EOF) {
	switch (o) {
	  case 'k':
	    crossval = TRUE;
	    folds = atoi(optarg); 
	    break;
	  case 'l':
	    crossval = TRUE;
	    folds = 0;
	    break;
	  case 's':
	    use_stdin = TRUE;
	    break;
	  case 'R':
	    Predict = TRUE;
	    Predict_probabilities = TRUE;
	    break;
	  case 'X':
	    Transform_data = TRUE;
	    break;
	  case 'p':
	    Predict = TRUE;
	    break;
	  case 'm':
	    if (optarg[0]=='a' || optarg[0]=='A') {
		Metric = ACCURACY;
	    } else {
		Metric = F1;
	    }
	  case 'L':
	    loss_ratio = atof(optarg);
	    FP_cost = 2.0*loss_ratio/(loss_ratio+1.0);
	    FN_cost = 2.0/(loss_ratio+1.0);
	    trace(SUMM) {
		printf("option: ratio of cost of FP to cost of FN is %g:%g\n",
		       FP_cost,FN_cost);
	    }
	    break;
	  case 'P':
	    Beta = atof(optarg);
	    break;
	  case 'N':
	    Gamma = atof(optarg);
	    break;
	  case '2':
	    Two_prototypes = TRUE;
	    break;
	  case 'v':
	    set_trace_level(atoi(optarg)); 
	    break;
	  case '?':
	  default: 
	    give_help();
	    if (o=='h') exit(0);
	    else fatal("option not implemented");
	}
    }
    test_data = NULL;
    if (optind<argc) {
	stem = argv[optind++];
	ld_names(add_ext(stem,".names"));
	if (use_stdin) train_data = ld_data(NULL);
	else {
	    train_data = ld_data(add_ext(stem,".data"));
	    test_data = ld_data(add_ext(stem,".test"));
	}
    } else {
	train_data = ld_data(NULL);
    }
    if (optind<argc) {
	warning("not all arguments were used: %s ...",argv[optind]);
    }

    if (!train_data || vmax(train_data)==0) fatal("no examples");
    if (!set_field(0) || n_fields()>1) {
	fatal("rocchio requires one set-valued field");
    }

    /* allocate internally used weight vectors */ 
    alloc_wts();

    if (crossval) {
	cross_validate(folds,train_data,&train_rocchio,&test_rocchio);
    } else {
	model_rocchio(train_data);

	err = rocchio_error_rate(train_data,stem,FALSE);
	trace(SUMM) {
	    printf("Train error rate:   %.2f%% +/- %.2f%% (%d datapoints)   <<\n",
		   100*err,
		   100*binomial_std_err(err,vmax(train_data)),
		   vmax(train_data));
	}

	if (test_data) {
	    err = rocchio_error_rate(test_data,stem,Predict);
	    trace(SUMM) {
		printf("Test error rate:   %.2f%% +/- %.2f%% (%d datapoints)   <<\n",
		       100*err,
		       100*binomial_std_err(err,vmax(test_data)),
		       vmax(test_data));
	    }
	}
	
	if (Transform_data) {
	    if (stem==NULL) stem="foo";
	    if ((fptrain = fopen(add_ext(stem,"_cos.data"),"w"))==NULL) {
		fatal("can't open file for transformed training data");
	    }
	    if (test_data && 
		(fptest = fopen(add_ext(stem,"_cos.test"),"w"))==NULL) {
		fatal("can't open file for transformed test data");
	    }
	    for (i=0; i<vmax(train_data); i++) {
		exi = vref(example_t,train_data,i);
		if (Two_prototypes) {
		    fprintf(fptrain,"%g,%g,",
			   inner_product(words(exi),Sum_wt_p),
			   inner_product(words(exi),Sum_wt_n));
		} else {
		    fprintf(fptrain,"%g,",eval_rule(words(exi)));
		}
		fprint_example(fptrain,exi);
	    }
	    if (test_data) {
		for (i=0; i<vmax(test_data); i++) {
		    exi = vref(example_t,test_data,i);
		    if (Two_prototypes) {
			fprintf(fptest,"%g,%g,",
				inner_product(words(exi),Sum_wt_p),
				inner_product(words(exi),Sum_wt_n));
		    } else {
			fprintf(fptest,"%g,",eval_rule(words(exi)));
		    }
		    fprint_example(fptest,exi);
		}
	    } /* if test data */
	} /* if transform data */
    } /* else not cross validating */
}

static double binomial_std_err(double err_rate,int n)
{
    return sqrt( (err_rate)*(1-err_rate)/((double)n-1) );
}

/* compute error rate and maybe print predictions */
static double rocchio_error_rate(vec_t *data,char *stem,BOOL print_predictions)
{
    int i;
    example_t *exi;
    symbol_t *pos_cl,*neg_cl;
    symbol_t *actual,*predicted;
    ex_count_t err,tot;
    double p;
    FILE *fp=NULL;

    if (print_predictions) {
	fp = fopen(add_ext(stem,".pred"),"w");
	if (fp==NULL) 
	  error("can't open prediction file %s",add_ext(stem,".pred"));
    }

    pos_cl = vref(atom_t,Classes,0)->nom;
    neg_cl = vref(atom_t,Classes,1)->nom;
    trace (LONG) 
      printf("// evaluating weights on %d examples\n",vmax(data)); 
    
    err = tot = 0;
    for (i=0; i<vmax(data); i++) {
	exi = vref(example_t,data,i);
	tot += exi->wt;
	actual = exi->lab.nom;
	predicted = eval_final_rule(words(exi)) ? pos_cl : neg_cl;
	if (actual!=predicted) err += exi->wt;
	if (fp!=NULL) {
	    if (Predict_probabilities) {
		p = eval_rule(words(exi))/Max_sim;
		fprint_symbol(fp,actual);
		fprintf(fp," %.20f ",p);
		fprint_symbol(fp,predicted);
		fprintf(fp,"\n");
	    } else {
		fprint_symbol(fp,predicted);
		fprintf(fp," 1 1 ");
		fprint_symbol(fp,actual);
		fprintf(fp,"\n");
	    }
	}
    }
    if (fp!=NULL) fclose(fp);
    return err/tot;
}

/****************************************************************************/

/* implementation of this algorithm stolen wholesale from Rob */

#define NUM_THRESH_INC  100
#define MAX_LOG_PRECOMP    (256)
#ifndef MAXDOUBLE
#define MAXDOUBLE MAXREAL
#endif

static int *Seen;
static double *Wt;
static double *Feature_wt;
static double Log_precomp[MAX_LOG_PRECOMP];
static double ltc_wt(int, int);
static void init_wts(symbol_t *,vec_t *);
static void find_threshold(symbol_t *,vec_t *);
static void finalize_wts(symbol_t *,vec_t *);
static double Final_thresh; 

/************
** classify an instance
************/
static int eval_final_rule(vec_t *inst)
{
    if (Two_prototypes) {
	return (inner_product(inst,Sum_wt_p) >= inner_product(inst,Sum_wt_n));
    } else {
	return (eval_rule(inst) >= Final_thresh);
    }
}


/************
** computes similarity of an instance to the prototype for pos
************/
static double eval_rule(vec_t *inst) 
{
    return inner_product(inst,Sum_wt);
}

/************
** computes similarity of an instance a given prototype vector
************/
static double inner_product(vec_t *inst,double *sum_wt)
{
  int k,aix;
  double sum, sq_sum;
  double w;
  

  /* if inst is word wi, set Seen[k] to f_ik */
  for (k=0; k<vmax(inst); k++) {
      aix = (*vref(symbol_t *,inst,k))->index;
      Seen[aix]++;
  }
  /* compute sum_{k=1}^t (w_{ik} * w_{+k}) 
                    w_{+k} * tf_{ik} * log(N_D/n_k)
     = sum_{k=1}^t           ----------------------
		    sqrt(sum_{j=1}^t [tf_{ij} * log(N_D/n_k)]^2  

     sq_sum is the thing inside sqrt, Sum_wt[k] is w_{+k}
     on iteration j,  w is tf_{ik} * log(N_D/n_k)
  */
  sum = sq_sum = 0.0;
  for (k=0; k<vmax(inst); k++) {
      aix = (*vref(symbol_t *,inst,k))->index;
      if (Seen[aix]) {
	  w = ltc_wt(Seen[aix], aix);
	  sq_sum += w*w;
	  sum += w*sum_wt[aix];
	  Seen[aix] = 0;
      }
  }

  if (sum==0.0) return 0;
  else return sum/sqrt(sq_sum);
}


/****************************************************************************/

/********************
** top-level learning algorithm
**
** see description in Ittner, Lewis, Ahn 94
**
********************/

static void model_rocchio(vec_t *data)
{
  int i,j;
  double m;
  double d;
  double sq_sum;
  example_t *exi;
  int ajx;
  symbol_t *pos_cl;
  ex_count_t p,n;
  atom_t *c0,*c1;
  vec_t *seti;

  /* count and order the classes--pos_cl is minority class */
  if (vmax(Classes)!=2) {
      fatal("this implementation of rocchio only handles two-class problems");
  }
  count_classes(data);
  reorder_classes(data);
  pos_cl = vref(atom_t,Classes,0)->nom;
  p = Class_counts[0];
  n = Class_counts[1];
  trace(SUMM) {
      printf("// %g examples (%g/%g)\n",p+n,p,n);
      printf("// class '%s' is treated as positive\n",pos_cl->name);
      fflush(stdout);
  }

  /* initialize the weights---
     for each word index i, initialize Seen[i] to zero,
     Sum_wt[i] to zero, and set Feature_wt[i] to $N_D/n_k$
  */   
  trace(SUMM) {
      printf("// initializing weights...\n");
      fflush(stdout);
  }
  init_wts(pos_cl,data);

  trace(SUMM) {
      printf("// setting weights...\n");
      fflush(stdout);
  }

  /* for each word i set Sum_wt[j] to w'_{+j} 
     where w_{+j} = beta/|R_+|  * sum_{i in R_+} w_ij - 
                    gamma/|R_-| * sum_{i in R_-} w_ij
  */		    

  for (i=0; i<vmax(data); i++) {
      exi = vref(example_t,data,i);
      seti = words(exi);
      /* count occurances of each word j in document i 
         ie Seen[i] is set to fik */
      for (j=0; j<vmax(seti); j++) {
	  ajx = (*vref(symbol_t *,seti,j))->index;
	  Seen[ajx]++;
      }
      /* set Wt[j] = tf_{ij} log(N_D/N) 
	 and sq_sum to sum_{j=1}^t [ tf_{ij} log(N_D/N) ]^2 
       */
      sq_sum = 0.0;
      for (j=0; j<vmax(seti); j++) {
	  ajx = (*vref(symbol_t *,seti,j))->index;
	  if (Seen[ajx] > 0) {
	      Wt[ajx] = ltc_wt(Seen[ajx], ajx);
	      sq_sum += Wt[ajx]*Wt[ajx];
	      Seen[ajx] = -1;   /* only update for each word once */
	  }
      }
      /* increment Sum_wt appropriately */
      if (Two_prototypes) {
	  m = (exi->lab.nom==pos_cl 
	       ? 1.0/ (p * sqrt(sq_sum))
	       : 1.0/ (n * sqrt(sq_sum)) );
      } else {
	  m = (exi->lab.nom==pos_cl 
	       ? Beta / (p * sqrt(sq_sum))
	       : -Gamma / (n * sqrt(sq_sum)) );
      }
      for (j=0; j<vmax(seti); j++) {
	  ajx = (*vref(symbol_t *,seti,j))->index;
	  if (Seen[ajx]) {
	      Sum_wt[ajx] += m * Wt[ajx];
	      if (Two_prototypes) {
		  if (exi->lab.nom==pos_cl) {
		      Sum_wt_p[ajx] += m * Wt[ajx];		      
		  } else {
		      Sum_wt_n[ajx] += m * Wt[ajx];		      
		  }
	      }
	      Seen[ajx] = 0;
	  }
      }
  }
  trace(SUMM) {
      printf("// finalizing weights...\n");
      fflush(stdout);
  }

  /* sets Sum_wt[i] to w_{+k} and finds threshold */
  finalize_wts(pos_cl,data);
}


/***********
** computes (log(f)+1)*log(number docs/number docs with feature i)
***********/
static double ltc_wt(int f, int i)
{
  return (f < MAX_LOG_PRECOMP ? Log_precomp[f] : 1.0 + log(f)) *
    Feature_wt[i];
}

/***********
** Allocates the weight vectors 
***********/

static void alloc_wts()
{
  Seen = newmem(n_symbolic_values(),int);
  Wt = newmem(n_symbolic_values(),double);
  Sum_wt = newmem(n_symbolic_values(),double);
  if (Two_prototypes) {
      Sum_wt_p = newmem(n_symbolic_values(),double);
      Sum_wt_n = newmem(n_symbolic_values(),double);
  }
  Feature_wt = newmem(n_symbolic_values(),double);
}


/***********
** Initializes the weights of the weight vector.
***********/
static void init_wts(symbol_t *pos_cl,vec_t *data)
{
  int i;
  ex_count_t p,n;
  
  for (i=0; i<n_symbolic_values(); i++)  {
      Seen[i] = 0;
      Wt[i] = Feature_wt[i]  = 0.0;
      Sum_wt[i] = 0.0;
      if (Two_prototypes) {
	  Sum_wt_p[i] = Sum_wt_n[i] = 0.0;
      }
  }

  compute_field_stats(pos_cl,0,data);
  for (i = 0; i < n_visited_symbols(); i++) {
      pos_field_stat(visited_symbol(i),&p,&n);
      Seen[i] = 0;
      Sum_wt[i] = 0.0;
      if (Two_prototypes) {
	  Sum_wt_p[i] = Sum_wt_n[i] = 0.0;
      }
      if (p+n > 0) {
	  Feature_wt[i] = log(((double) vmax(data))/(p+n));
      }
  }
  Log_precomp[0] = 0.0;
  for (i = 1; i < MAX_LOG_PRECOMP; i++)
    Log_precomp[i] = 1.0 + log(i);
}

/***********
** Finalizes weights for use by final evaluation rule.
***********/
static void finalize_wts(symbol_t *pos_cl,vec_t *data)
{
  int i;

  for (i = 0; i < n_symbolic_values(); i++) {
      if (Sum_wt[i] < 0.0) {
	  Sum_wt[i] = 0.0;
      }
      if (Two_prototypes) {
	  if (Sum_wt_p[i] < 0.0) {
	      Sum_wt_p[i] = 0.0;
	  }
	  if (Sum_wt_n[i] < 0.0) {
	      Sum_wt_n[i] = 0.0;
	  }
      }
  }
  find_threshold(pos_cl,data);
}

/************
** finds best threshold for weight function 
************/

typedef struct pair_s {
    double val;
    BOOL lab;
} pair_t;

static int compare_pair(pi,pj)
pair_t *pi,*pj;
{
    if (pi->val > pj->val) return 1;
    else if (pi->val == pj->val) return 0;
    else return -1;
}

static void find_threshold(symbol_t *pos_cl,vec_t *data)
{
  int t;
  example_t *ext;
  pair_t *pair;
  double bestf,besty,f;
  double y;
  int c[2][2],n_pos,n_neg;
  
  /* allocate and sort pairs, and count # pos/neg examples */
  n_pos = n_neg = 0;
  pair = newmem(vmax(data),pair_t);
  for (t=0; t<vmax(data); t++) {
      ext = vref(example_t,data,t);
      pair[t].val = eval_rule(words(ext));
      pair[t].lab = (ext->lab.nom == pos_cl);
      if (pair[t].lab) n_pos++;
      else n_neg++;
  }
  qsort((char *)pair,vmax(data),sizeof(pair_t),&compare_pair);

  /* find greatest similarity value for training data */
  Max_sim = pair[vmax(data)-1].val;

  trace(DBG1) {
      printf("// sorted pairs:\n");
      for (t=0; t<vmax(data); t++) {
	  printf("// %3d: (%g,%c)\n", t,pair[t].val,"-+"[pair[t].lab]);
      }
  }

  bestf = -MAXREAL;
  /* c[act][pred] will initially be confusion matrix for 
     lowest possible threshold---ie all predicted pos */
  c[0][0] = 0;                /* # neg < threshold y */
  c[1][0] = 0;                /* # pos < threshold y */
  c[0][1] = n_neg;            /* # neg >= threshold y */
  c[1][1] = n_pos;            /* # pos >= threshold y */

  for (t=0; t<vmax(data); ) {
      y = pair[t].val;
      /* see if current confusion matrix is best */
      if (Metric==F1) {
	  f = (2.0*c[1][1])/(2*c[1][1]+c[1][0]+c[0][1]);
      }
      else if (Metric==ACCURACY) {
	  f = -(FP_cost * c[0][1]) - (FN_cost * c[1][0]);
      }
      else fatal("unknown optimization metric!");
      if (f>bestf) {
	  if (t==0) besty = y;
	  else besty = (y + pair[t-1].val)/2.0;
	  bestf = f;
	  trace(DBUG) {
	      printf("// use=T ");
	  }
      } else {
	  trace(DBUG) {
	      printf("// use=F ");
	  }
      }
      /* update confusion matrix for next threshold, and advance 
	 t to point to that new threshold (with val!=y) 
       */
      while (t<vmax(data) && pair[t].val==y) {
	  c[pair[t].lab][0]++;
	  c[pair[t].lab][1]--;
	  trace(DBUG) {
	      printf("%3d: (%g,%c)\tv=%g\t00=%d 01=%d 10=%d 11=%d\n", 
		     t,pair[t].val,"-+"[pair[t].lab],f,
		     c[0][0],c[0][1],c[1][0],c[1][1]);
	  }
	  t++;
      }
  }
  /* also consider the default concept, which rejects everything */
  if (vmax(data)>0) {
      /* make a value bigger than the max value */
      y = (pair[vmax(data)-1].val+1.0) * 1.1;
      if (Metric==F1) {
	  f = (2.0*c[1][1])/(2*c[1][1]+c[1][0]+c[0][1]);
      }
      else if (Metric==ACCURACY) {
	  f = -(FP_cost * c[0][1]) - (FN_cost * c[1][0]);
      }
      else fatal("unknown optimization metric!");
      if (f>=bestf) {
	  besty = y;
	  bestf = f;
	  trace(DBUG) {
	      printf("// using [all neg] threshold=%g (value %f)\n",
		     besty,bestf);
	  }
      }
  }      
  trace(LONG) {
      printf("// best threshold=%g (value %f)\n",besty,bestf);
  }
  Final_thresh = besty;
  freemem(pair);
}
